/* ---------------------------------------------------------------------------

                 Nullsoft Database Engine - Codename: Neutrino

--------------------------------------------------------------------------- */

/* ---------------------------------------------------------------------------

 Global Include File

--------------------------------------------------------------------------- */

#ifndef __NDE_H
#define __NDE_H

extern char *tSign;
extern char *iSign;

// Magic headers
#define __TABLE_SIGNATURE__ tSign
#define __INDEX_SIGNATURE__ iSign

// Linked list entry types
#define UNKNOWN           0
#define FIELD             1
#define FILTER           2
#define SCANNER           3

// Field types
#define FIELD_CLONE     255

#define FIELD_UNKNOWN   255
#define FIELD_COLUMN     0
#define FIELD_INDEX       1
#define FIELD_REDIRECTOR 2
#define FIELD_STRING     3
#define FIELD_INTEGER     4
#define FIELD_BOOLEAN     5
#define FIELD_BINARY     6
#define FIELD_GUID       7
#define FIELD_PRIVATE     8
#define FIELD_BITMAP     6
#define FIELD_FLOAT       9
#define FIELD_DATETIME   10
#define FIELD_LENGTH     11

#define DELETABLE         1
#define NONDELETABLE     2

// Records constants
#define NEW_RECORD    -128
#define FIELD_NOT_FOUND -1
#define INVALID_RECORD  -1

// Index constants
#define PRIMARY_INDEX   255
#define QFIND_ENTIRE_SCOPE -1
#define FIRST_RECORD    -1

#define throwException(x) {}

// Filter types
enum {
  FILTER_NONE = 100,
  FILTER_EQUALS,
  FILTER_NOTEQUALS,
  FILTER_CONTAINS,
  FILTER_NOTCONTAINS,
  FILTER_ABOVE,
  FILTER_BELOW,
  FILTER_ABOVEOREQUAL,
  FILTER_BELOWOREQUAL,
  FILTER_BEGINS,
  FILTER_ENDS,
  FILTER_LIKE,
  FILTER_ISEMPTY,
  FILTER_ISNOTEMPTY,
};

#define FILTER_NOT       0
#define FILTER_AND       1
#define FILTER_OR         2

#define FILTERS_INVALID    -1
#define FILTERS_INCOMPLETE  1
#define FILTERS_COMPLETE    1
#define ADDFILTER_FAILED    2

// Permissions
#define PERM_DENYALL     0
#define PERM_READ         1
#define PERM_READWRITE   3

// All our classes so we can foreward reference
class LinkedListEntry;
class LinkedList;
class Field;
class ColumnField;
class IndexField;
class StringField;
class IntegerField;
class GUIDField;
class FloatField;
class BooleanField;
class BinaryField;
class BitmapField;
class PrivateField;
class Record;
class Scanner;
class Table;
class Index;
class Filter;
class Database;
class NdeApi;

#ifdef NDE_UNIX
#define NDE_NOWIN32FILEIO
#define NO_TABLE_WIN32_LOCKING
typedef int BOOL;
typedef void* HANDLE;
#define TRUE 1
#define FALSE 0
typedef struct _GUID {
  unsigned long Data1;
  unsigned short Data2;
  unsigned short Data3;
  unsigned char Data4[8];
} GUID;
#define HINSTANCE int
#define HWND int
#define DWORD int
#define wsprintf sprintf
#define OutputDebugString(x) ;
#define GetCurrentProcessId() getpid()
#include <stdlib.h>
#include <ctype.h>
#include <malloc.h>
#include <strings.h>
#include <string.h>
#include <sys/stat.h>
#define stricmp strcasecmp
#define strcmpi strcasecmp
#define strnicmp strncasecmp
#define strncmpi strncasecmp
void clear_stdin(void);
#define _MAX_PATH 8192
#define _MAX_DRIVE 256
#define _MAX_DIR 7424
#define _MAX_FNAME 256
#define _MAX_EXT 256
#include "splitpath.h"
void CopyFile(const char *filename, const char *destfilename, BOOL b);
void DeleteFile(const char *filename);
BOOL MoveFile(const char *filename, const char *destfilename);
void Sleep(int ms);
#define min(x,y) ((x > y) ? y : x)
#define max(x,y) ((x < y) ? y : x)
#endif

// All our includes
#include "Vfs.h"
#include "LinkedList.h"
#include "Field.h"
#include "ColumnField.h"
#include "IndexField.h"
#include "StringField.h"
#include "IntegerField.h"
#include "Record.h"
#include "Scanner.h"
#include "Table.h"
#include "Database.h"
#include "Index.h"
#include "Filter.h"
#include "DBUtils.h"

#endif
