#include "NDE.h"

typedef struct {
  char *token;
  int tid;
} tokenstruct;

tokenstruct Tokens[] = { // Feel free to add more...
  {"AND", TOKEN_AND },
  {"OR", TOKEN_OR },
  {"HAS", TOKEN_CONTAINS },
  {"NOTHAS",TOKEN_NOTCONTAINS},
  {"BEGINS", TOKEN_BEGINS },
  {"ENDS", TOKEN_ENDS },
  {"ISEMPTY", TOKEN_ISEMPTY},
  {"ISNOTEMPTY",TOKEN_ISNOTEMPTY},
  {"LIKE", TOKEN_LIKE},
};

typedef struct {
  int Op;
  int Level;
} OpLevel;

/*
    our state machine

                               &, |
            ----------<-------------------------<----------------------<-----------
           |                                                                       |
           v    ID (Col)        =, >, <...       ID / Data / [f]          )        |
     ---->(0) ----->-----> (1) ------>-----> (2) ------>------> (3) ------>-----> (4) <--
    |     |^                \isempty------------->------------/  |^                |     |
    | !(  ||                                                     ||----            |  )  |
     --<--  ---------<---------------------------<-------------<-|     |            -->--
                               &, |                              v [f] |
                                                                  -->--
                                                                   
*/

//---------------------------------------------------------------------------
BOOL Scanner::Query_Parse(char *query) {
  char *p = query; // pointer on next token to read
  int size;
  int state = 0;
  int pcount = 0;
  VListEntry<OpLevel> *entry; 

  if (pstack->GetNElements() > 0)
    Query_CleanUp();
  

  while (1) {
    p = Query_EatSpace(p);
    int t = Query_GetNextToken(p, &size, &token);
    if (t == TOKEN_UNKNOWN) {
      Query_SyntaxError(p-query);
      return FALSE;
    }
    if (t == TOKEN_EOQ)
      break;
    switch (state) {
      case 0:
        switch (t) {
          case TOKEN_PAROPEN:
            state = 0;
            // check too many parenthesis open
            if (pcount == 255) {
              Query_SyntaxError(p-query); // should not be _syntax_ error
              return FALSE;
            }
            // up one level
            pcount++;
            break;
          case TOKEN_NOT:
            // push not in this level
            OpLevel o;
            o.Op = FILTER_NOT;
            o.Level = pcount;
            entry = new VListEntry<OpLevel>;
            entry->SetVal(o);
            pstack->AddEntry(entry, TRUE);
            state = 0;
            break;
          case TOKEN_IDENTIFIER:
            state = 1;
            // create filter column
            if (AddFilterByName(token, NULL, FILTER_NONE) == ADDFILTER_FAILED) {
              Query_SyntaxError(p-query);
              return FALSE;
            }
            break;
          default:
            Query_SyntaxError(p-query);
            return FALSE;
        }
        break;
      case 1:
        switch (t) {
          case TOKEN_EQUAL: {
            state = 2;
            // set filter op
            Filter *f = GetLastFilter();
            f->SetOp(FILTER_EQUALS);
            break;
            }
          case TOKEN_ABOVE: {
            state = 2;
            // set filter op
            Filter *f = GetLastFilter();
            f->SetOp(FILTER_ABOVE);
            break;
            }
          case TOKEN_BELOW: {
            state = 2;
            // set filter op
            Filter *f = GetLastFilter();
            f->SetOp(FILTER_BELOW);
            break;
            }
          case TOKEN_CONTAINS: {
            state = 2;
            // set filter op
            Filter *f = GetLastFilter();
            f->SetOp(FILTER_CONTAINS);
            break;
            }
          case TOKEN_NOTCONTAINS: {
            state = 2;
            // set filter op
            Filter *f = GetLastFilter();
            f->SetOp(FILTER_NOTCONTAINS);
            break;
            }
          case TOKEN_AOREQUAL: {
            state = 2;
            // set filter op
            Filter *f = GetLastFilter();
            f->SetOp(FILTER_ABOVEOREQUAL);
            break;
            }
          case TOKEN_BOREQUAL: {
            state = 2;
            // set filter op
            Filter *f = GetLastFilter();
            f->SetOp(FILTER_BELOWOREQUAL);
            break;
            }
          case TOKEN_NOTEQUAL: {
            state = 2;
            // set filter op
            Filter *f = GetLastFilter();
            f->SetOp(FILTER_NOTEQUALS);
            break;
          }
          case TOKEN_BEGINS: {
            state = 2;
            Filter *f = GetLastFilter();
            f->SetOp(FILTER_BEGINS);
          }
          break;
          case TOKEN_ENDS: {
            state = 2;
            Filter *f = GetLastFilter();
            f->SetOp(FILTER_ENDS);
          }
          break;
          case TOKEN_LIKE: {
            state = 2;
            // set filter op
            Filter *f = GetLastFilter();
            f->SetOp(FILTER_LIKE);
            break;
            }
          case TOKEN_ISNOTEMPTY:
          case TOKEN_ISEMPTY: {
            state = 3;
            Filter *f = GetLastFilter();
            f->SetOp(t==TOKEN_ISEMPTY ? FILTER_ISEMPTY : FILTER_ISNOTEMPTY);
            // pop all operators in this level beginning by the last inserted
            entry = (VListEntry<OpLevel> *)pstack->GetFoot();
            while (entry) {
              if (entry->GetVal().Level == pcount) {
                AddFilterOp(entry->GetVal().Op);
                VListEntry<OpLevel> *_entry = (VListEntry<OpLevel> *)entry->GetPrevious(); 
                pstack->RemoveEntry(entry);
                entry = _entry;
              } else 
                break;
            }
          }
          break;
          default:
            Query_SyntaxError(p-query);            
            return FALSE;
        }
        break;
      case 2:
          if (t == TOKEN_SQBRACKETOPEN) {
            state = 3;
            char *s = strchr(p, ']');
            if (!s) {
              Query_SyntaxError(p-query);
              return FALSE;
            }
            p = Query_EatSpace(p);
            if (*p == '[') p++;
            char *format = (char *)malloc(s-p+1);
            strncpy(format, p, s-p);
            format[s-p] = 0;

            Filter *f = GetLastFilter();
            int id = f->GetId();
            ColumnField *c = GetColumnById(id);
            int tt = c ? c->GetDataType() : -1;
            if (disable_date_resolution || !c || (tt != FIELD_INTEGER && tt != FIELD_DATETIME && tt != FIELD_LENGTH && tt != FIELD_BOOLEAN)) {

              if (disable_date_resolution)
              {
                StringField *field = (StringField *)f->Data();
            
                if (!field) {
                  // format was used without a value, assume value is 0
                  f->SetData(new StringField(""));
                  entry = (VListEntry<OpLevel> *)pstack->GetFoot();
                  while (entry) {
                    if (entry->GetVal().Level == pcount) {
                      AddFilterOp(entry->GetVal().Op);
                      VListEntry<OpLevel> *_entry = (VListEntry<OpLevel> *)entry->GetPrevious(); 
                      pstack->RemoveEntry(entry);
                      entry = _entry;
                    } else 
                      break;
                  }
                  field = (StringField*)f->Data();
                }
            
                f->SetInfoText(format);
                field->SetString(format);
                free(format);
                p = s+1;
                continue;
              }
              

              free(format);
              Query_SyntaxError(p-query);
              return FALSE;
            }

            IntegerField *field = (IntegerField *)f->Data();
            
            if (!field) {
              // format was used without a value, assume value is 0
              f->SetData(new IntegerField(0));
              entry = (VListEntry<OpLevel> *)pstack->GetFoot();
              while (entry) {
                if (entry->GetVal().Level == pcount) {
                  AddFilterOp(entry->GetVal().Op);
                  VListEntry<OpLevel> *_entry = (VListEntry<OpLevel> *)entry->GetPrevious(); 
                  pstack->RemoveEntry(entry);
                  entry = _entry;
                } else 
                  break;
              }
              field = (IntegerField *)f->Data();
            }
            
            f->SetInfoText(format);
            int r = field->ApplyConversion(format);
            free(format);
            if (!r) {
              Query_SyntaxError(p-query);
              return FALSE;
            }
            p = s+1;
            continue;
          }
//        switch (t) {
  //      case TOKEN_IDENTIFIER: 
        else { // JF> we make this relaxed, so anything is valid as a value
            state = 3;
            // set filter data
            Filter *f = GetLastFilter();
            int id = f->GetId();
            ColumnField *c = GetColumnById(id);
            switch (c ? c->GetDataType() : -1) {
              case FIELD_DATETIME: 
                if (disable_date_resolution)
                  goto field_string_override;
              case FIELD_LENGTH: 
                {
                  int i;
                  IntegerField *i_f = new IntegerField();
                  i = atoi(token); // todo: Replace with own conversion and error checking
                  char *p;
                  if ((p=strstr(token,":")))
                  {
                    i*=60;
                    i+=atoi(++p);
                    if ((p=strstr(p,":")))
                    {
                      i*=60;
                      i+=atoi(++p);
                    }
                  }
                  i_f->SetValue(i);
                  f->SetData(i_f);
                }
              break;

              case FIELD_BOOLEAN:
              case FIELD_INTEGER: 
                {
                  int i;
                  IntegerField *i_f = new IntegerField();
                  i = atoi(token); // todo: Replace with own conversion and error checking
                  i_f->SetValue(i);
                  f->SetData(i_f);
                }
              break;
              case FIELD_STRING: 
        field_string_override:
                {
                StringField *s_f = new StringField();
                s_f->SetString(token);
                f->SetData(s_f);
                }
                break;
              default:
                Query_SyntaxError(p-query);
                return FALSE;
                break;

            }
            // pop all operators in this level beginning by the last inserted
            entry = (VListEntry<OpLevel> *)pstack->GetFoot();
            while (entry) {
              if (entry->GetVal().Level == pcount) {
                AddFilterOp(entry->GetVal().Op);
                VListEntry<OpLevel> *_entry = (VListEntry<OpLevel> *)entry->GetPrevious(); 
                pstack->RemoveEntry(entry);
                entry = _entry;
              } else 
                break;
            }
            break;
            }
//          default: 
  //          Query_SyntaxError(p-query);
    //        return FALSE;
//        }
        break;
      case 3:
        switch (t) {
          case TOKEN_SQBRACKETOPEN: {
            char *s = strchr(p, ']');
            if (!s) {
              Query_SyntaxError(p-query);
              return FALSE;
            }
            p = Query_EatSpace(p);
            if (*p == '[') p++;
            char *format = (char *)malloc(s-p+1);
            strncpy(format, p, s-p);
            format[s-p] = 0;

            Filter *f = GetLastFilter();
            int id = f->GetId();
            ColumnField *c = GetColumnById(id);
            int tt = c ? c->GetDataType() : -1;
            if (disable_date_resolution || !c || (tt != FIELD_INTEGER && tt != FIELD_DATETIME && tt != FIELD_LENGTH && tt != FIELD_BOOLEAN)) {
              if (disable_date_resolution)
              {
                StringField *field = (StringField *)f->Data();
            
                if (!field) {
                  // format was used without a value, assume value is 0
                  f->SetData(new StringField(""));
                  entry = (VListEntry<OpLevel> *)pstack->GetFoot();
                  while (entry) {
                    if (entry->GetVal().Level == pcount) {
                      AddFilterOp(entry->GetVal().Op);
                      VListEntry<OpLevel> *_entry = (VListEntry<OpLevel> *)entry->GetPrevious(); 
                      pstack->RemoveEntry(entry);
                      entry = _entry;
                    } else 
                      break;
                  }
                  field = (StringField *)f->Data();
                }
            
                f->SetInfoText(format);
                field->SetString(format);
                free(format);
                p = s+1;
                continue;
              }
              free(format);
              Query_SyntaxError(p-query);
              return FALSE;
            }

            IntegerField *field = (IntegerField *)f->Data();
            
            if (!field) {
              // format was used without a value, assume value is 0
              f->SetData(new IntegerField(0));
              entry = (VListEntry<OpLevel> *)pstack->GetFoot();
              while (entry) {
                if (entry->GetVal().Level == pcount) {
                  AddFilterOp(entry->GetVal().Op);
                  VListEntry<OpLevel> *_entry = (VListEntry<OpLevel> *)entry->GetPrevious(); 
                  pstack->RemoveEntry(entry);
                  entry = _entry;
                } else 
                  break;
              }
              field = (IntegerField *)f->Data();
            }
            
            f->SetInfoText(format);
            int r = field->ApplyConversion(format);
            free(format);
            if (!r) {
              Query_SyntaxError(p-query);
              return FALSE;
            }
            p = s+1;
            continue;
          }
          break;

          case TOKEN_PARCLOSE:
            state = 4;
            // check parenthesis count
            if (pcount == 0) {
              Query_SyntaxError(p-query);
              return FALSE;
            }
            // down one level
            pcount--;
            // pop all operators in this level, beginning by the last inserted
            while (entry) {
              if (entry->GetVal().Level == pcount) {
                AddFilterOp(entry->GetVal().Op);
                VListEntry<OpLevel> *_entry = (VListEntry<OpLevel> *)entry->GetPrevious(); 
                pstack->RemoveEntry(entry);
                entry = _entry;
              } else 
                break;
            }
            break;
          case TOKEN_AND: {
            state = 0;
            // push and
            OpLevel o;
            o.Op = FILTER_AND;
            o.Level = pcount;
            entry = new VListEntry<OpLevel>;
            entry->SetVal(o);
            pstack->AddEntry(entry, TRUE);
            break;
            }
          case TOKEN_OR: {
            state = 0;
            // push or
            OpLevel o;
            o.Op = FILTER_OR;
            o.Level = pcount;
            entry = new VListEntry<OpLevel>;
            entry->SetVal(o);
            pstack->AddEntry(entry, TRUE);
            break;
            }
          default:
            Query_SyntaxError(p-query);
            return FALSE;
        }
        break;
      case 4:
        switch (t) {
          case TOKEN_AND: {
            state = 0;
            // push and
            OpLevel o;
            o.Op = FILTER_AND;
            o.Level = pcount;
            entry = new VListEntry<OpLevel>;
            entry->SetVal(o);
            pstack->AddEntry(entry, TRUE);
            break;
            }
          case TOKEN_OR: {
            state = 0;
            // push or
            OpLevel o;
            o.Op = FILTER_OR;
            o.Level = pcount;
            entry = new VListEntry<OpLevel>;
            entry->SetVal(o);
            pstack->AddEntry(entry, TRUE);
            break;
            }
          case TOKEN_PARCLOSE:
            state = 4;
            // check parenthesis count
            if (pcount == 0) {
              Query_SyntaxError(p-query);
              return FALSE;
            }
            // down one level
            pcount--;
            // pop all operators in this level, beginning by the last inserted
            while (entry) {
              if (entry->GetVal().Level == pcount) {
                AddFilterOp(entry->GetVal().Op);
                VListEntry<OpLevel> *_entry = (VListEntry<OpLevel> *)entry->GetPrevious(); 
                pstack->RemoveEntry(entry);
                entry = _entry;
              } else 
                break;
            }
            break;
          default:
            Query_SyntaxError(p-query);
            return FALSE;
        }
        break;
      default:
        // Ahem... :/
        break;
    }
  p += size;
  }
  if (pcount > 0) {
    Query_SyntaxError(p-query);
    return FALSE;
  }
  return TRUE;
}

//---------------------------------------------------------------------------
void Scanner::Query_SyntaxError(int c) {

}

//---------------------------------------------------------------------------
void Scanner::Query_CleanUp() {
  while (pstack->GetNElements() > 0) {
    VListEntry<int> *e;
    e = (VListEntry<int> *)pstack->GetHead();
    pstack->RemoveEntry(e);
  }
}


//---------------------------------------------------------------------------
char *Scanner::Query_EatSpace(char *p) {
  while (*p && *p == ' ') p++;
  return p;
}

//---------------------------------------------------------------------------
char *Scanner::Query_ProbeNonAlphaNum(char *p) {
int inquote=0;
  while (*p && (!Query_isControlChar(*p) || (inquote))) {
    if (*p == '\"') {
      if (!inquote)
        inquote = 1;
      else
        return p+1;
    }
    p++;
  }
  return p;
}

//---------------------------------------------------------------------------
int Scanner::Query_isControlChar(char p) {
  switch (p) {
    case '&':
    case '|':
    case '!':
    case '(':
    case '[':
    case ')':
    case ']':
    case '>':
    case '<':
    case '=':
    case ',':
    case ' ':
      return TRUE;
  }
  return FALSE;
}

//---------------------------------------------------------------------------
char *Scanner::Query_ProbeAlphaNum(char *p) {
  while (*p && Query_isControlChar(*p)) p++;
  return p;
}

//---------------------------------------------------------------------------
char *Scanner::Query_ProbeSpace(char *p) {
  while (*p && *p != ' ') p++;
  return p;
}

//---------------------------------------------------------------------------
int Scanner::Query_LookupToken(char *t) {
  for (int i=0;i<sizeof(Tokens)/sizeof(tokenstruct);i++) {
    if (!stricmp(Tokens[i].token, t))
      return Tokens[i].tid;
  }
  return TOKEN_IDENTIFIER;
}

//---------------------------------------------------------------------------
int Scanner::Query_GetNextToken(char *p, int *size, char **_token, int tokentable) {

  int t = TOKEN_EOQ;
  char *startptr = p;

  if (!*p) return TOKEN_EOQ;

  p = Query_EatSpace(p);

  char *e = Query_ProbeNonAlphaNum(p);

  if (e != p) { // We have a word
    if (*_token) free(*_token);
    *_token = strdup(p);
    (*_token)[e-p]=0; //token.rSplit(e-p);
    if (*(*_token) == '\"' && (*_token)[strlen(*_token)-1] == '\"') {
      int l=strlen(*_token)-2;
      if(l>0) {
        memcpy(*_token,(*_token)+1,l);
        (*_token)[l]=0;
        char *escaped = Query_Unescape(*_token);
        free(*_token);
        *_token = escaped;
      }
      else
        (*_token)[0]=0;// we have an empty string
    }

    switch (tokentable) {
      case -1:
        t = TOKEN_IDENTIFIER;
        break;
      case 0:
        t = Query_LookupToken(*_token);
        break;
      case 1:
        t = IntegerField::LookupToken(*_token);
    }
    p = e;
  } else { // We have a symbol
    switch (*p) {
      case '&':
        if (*(p+1) == '&') p++;
        t = TOKEN_AND;
        break;
      case '|':
        if (*(p+1) == '|') p++;
        t = TOKEN_OR;
        break;
      case '!':
        if (*(p+1) == '=') {
          p++;
          t = TOKEN_NOTEQUAL;
          break;
        }
        t = TOKEN_NOT;
        break;
      case '(':
        t = TOKEN_PAROPEN;
        break;
      case ')':
        t = TOKEN_PARCLOSE;
        break;
      case '[':
        t = TOKEN_SQBRACKETOPEN;
        break;
      case ']':
        t = TOKEN_SQBRACKETCLOSE;
        break;
      case ',':
        t = TOKEN_COMMA;
        break;
      case '>':
        if (*(p+1) == '=') {
          p++;
          t = TOKEN_AOREQUAL;
          break;
        }
        if (*(p+1) == '<') {
          p++;
          t = TOKEN_NOTEQUAL;
          break;
        }
        t = TOKEN_ABOVE;
        break;
      case '<':
        if (*(p+1) == '=') {
          p++;
          t = TOKEN_BOREQUAL;
          break;
        }
        if (*(p+1) == '>') {
          p++;
          t = TOKEN_NOTEQUAL;
          break;
        }
        t = TOKEN_BELOW;
        break;
      case '=':
        if (*(p+1) == '>') {
          p++;
          t = TOKEN_AOREQUAL;
          break;
        }
        if (*(p+1) == '<') {
          p++;
          t = TOKEN_BOREQUAL;
          break;
        }
        if (*(p+1) == '!') {
          p++;
          t = TOKEN_NOTEQUAL;
          break;
        }
        if (*(p+1) == '=') p++;
        t = TOKEN_EQUAL;
        break;
      default:
        t = TOKEN_UNKNOWN;
        break;
    }
  p++;
  }

  *size = p - startptr;
  return t;
}

// you must free the mem!
char *Scanner::Query_Unescape(char *p) {
  if (!p) return p;
  if (!*p) return strdup(p);
  int l = strlen(p);
  char *unescaped = (char *)malloc(l+1);
  char *d = unescaped;
  while (*p) {
    if (*p == '%') {
      int abort = 0;
      if (!*(p+1)) { *d++ = *p++; break; }
      if (*(p+1) == '%') {
        *d = *p;
        p = p+2;
        d++;
      }
      int a = *(p+1);
      int b = 0;
      if (a != 0) b = *(p+2);
      a = toupper(a);
      b = toupper(b);
      if (a >='0' && a <= '9') a -= '0';
      else if (a >= 'A' && a <= 'F') { a -= 'A'; a += 10; }
      else abort = 1;
      if (b >='0' && b <= '9') b -= '0';
      else if (b >= 'A' && b <= 'F') { b -= 'A'; b += 10; }
      else abort = 1;
      if (abort) {
        *d++ = *p++;
        continue;
      }
      *d++ = a*16+b;
      p = p+3;
      continue;
    }
    *d = *p;
    p++; d++;
  }
  *d = 0;
  return unescaped;
}

