/* ---------------------------------------------------------------------------
                           Nullsoft Database Engine
                             --------------------
                        codename: Near Death Experience
--------------------------------------------------------------------------- */

/* ---------------------------------------------------------------------------

 StringField Class

--------------------------------------------------------------------------- */

#include "NDE.h"

//---------------------------------------------------------------------------
StringField::StringField(const char *Str)
{
InitField();
Type = FIELD_STRING;
if (Str) String = strdup(Str);
Kind = DELETABLE;
}

//---------------------------------------------------------------------------
void StringField::InitField(void)
{
Type = FIELD_STRING;
String = NULL;
Kind = DELETABLE;
Physical = FALSE;
}

//---------------------------------------------------------------------------
StringField::StringField()
{
InitField();
}

//---------------------------------------------------------------------------
StringField::~StringField()
{
if (Kind != DELETABLE) throwException(EXCEPTION_NON_DELETABLE_OBJECT);
if (String) free(String);
}

//---------------------------------------------------------------------------
void StringField::ReadTypedData(unsigned char *data, int len)
{
unsigned short c;
int pos=0;

CHECK_SHORT(len);
c = GET_SHORT(); pos+=2;
if (c)
  {
  CHECK_BIN(len, c);
  String = (char *)malloc(c+1);
  GET_BINARY((unsigned char *)String, data, c, pos);
  String[c]=0;
  }
}

//---------------------------------------------------------------------------
void StringField::WriteTypedData(unsigned char *data, int len)
{
unsigned short c;
int pos=0;

if (String) c = (unsigned short)strlen(String)+1; else c = 0;
CHECK_SHORT(len);
PUT_SHORT(c); pos+=2;
if (String) 
  {
  CHECK_BIN(len, c);
  PUT_BINARY(data, (unsigned char*)String, c, pos);
  }
}

//---------------------------------------------------------------------------
char *StringField::GetString(void)
{
return String;
}

//---------------------------------------------------------------------------
void StringField::SetString(const char *Str)
{
if (Readonly) throwException(EXCEPTION_READ_ONLY_FIELD);
if (!Str) return;
if (Physical)
  {
  RememberSubtableEntry();
  free(String);
  String = NULL;
  }
String = strdup(Str);
}

//---------------------------------------------------------------------------
int StringField::GetDataSize(void)
{
if (!String) return 2;
return strlen(String)+3;
}

//---------------------------------------------------------------------------
int StringField::Compare(Field *Entry)
{
if (!Entry) return -1;
if (Entry->GetType() != GetType()) return 0;
return mystricmp(GetString(), ((StringField*)Entry)->GetString());
}

//---------------------------------------------------------------------------
BOOL StringField::ApplyFilter(Field *Data, int op)
{
BOOL r;
char *p = ((StringField *)Data)->GetString();
char *d = String;
if (!p)
  p = "";
if (!d)
  d = "";

switch (op)
  {
  case FILTER_EQUALS:
    r = (BOOL)!strcmpi(d, p);
    break;
  case FILTER_NOTEQUALS:
    r = (BOOL)strcmpi(d, p);
    break;
  case FILTER_CONTAINS:
    r = (BOOL)stristr(d, p);
    break;
  case FILTER_NOTCONTAINS:
    r = (BOOL)!stristr(d, p);
    break;
  case FILTER_ABOVE:
    r = (BOOL)(strcmpi(d, p) > 0);
    break;
  case FILTER_ABOVEOREQUAL:
    r = (BOOL)(strcmpi(d, p) >= 0);
    break;
  case FILTER_BELOW:
    r = (BOOL)(strcmpi(d, p) < 0);
    break;
  case FILTER_BELOWOREQUAL:
    r = (BOOL)(strcmpi(d, p) <= 0);
    break;
  case FILTER_BEGINS:
    r = (BOOL)(strnicmp(d, p, strlen(p)) == 0);
    break;
  case FILTER_ENDS: {
          int lenp = strlen(p), lend = strlen(d);
    if (lend < lenp) return 0;  // too short
    r = (BOOL)(strcmpi((d + lend) - lenp, p) == 0);
        }
    break;
  case FILTER_LIKE:
// todo: make configurable words to skip, as well as trailing whitespace removal
#define SKIP_THE_AND_WHITESPACE(x) { while (!isalnum(*x) && *x) x++; if (!strnicmp(x,"the ",4)) x+=4; while (*x == ' ') x++; }

    SKIP_THE_AND_WHITESPACE(d)
    SKIP_THE_AND_WHITESPACE(p)
    r = (BOOL)(stricmp(d, p) == 0);
    break;
  default:
    r = TRUE;
    break;
  }
return r;
}

