/* ---------------------------------------------------------------------------
                           Nullsoft Database Engine
                             --------------------
                        codename: Near Death Experience
--------------------------------------------------------------------------- */

/* ---------------------------------------------------------------------------

 Virtual File System

--------------------------------------------------------------------------- */

#ifdef NDE_NOWIN32FILEIO
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include "Vfs.h"

#ifndef EOF
#define EOF -1
#endif

//----------------------------------------------------------------------------
VFILE *Vfopen(char *fl, char *mode, BOOL Cached) {
  if (!fl) return NULL;
  VFILE *f = (VFILE *)malloc(sizeof(VFILE));
  if (!f)
    return NULL;
  f->mode = 0;
  f->data = NULL;
  f->filesize = 0;
  f->maxsize = 0;
  f->ptr = 0;
  f->dirty=0;
  f->flushtable=0;
#ifdef NDE_ALLOW_NONCACHED
  f->cached = Cached;
#endif
  if (!strchr(mode, '+'))  {
    if (strchr(mode, 'r'))
      f->mode = VFS_READ | VFS_MUSTEXIST;
    if (strchr(mode, 'w'))
      f->mode = VFS_WRITE | VFS_CREATE | VFS_NEWCONTENT;
    if (strchr(mode, 'a'))
      f->mode = VFS_WRITE | VFS_CREATE | VFS_SEEKEOF;
  } else {
    if (strstr(mode, "r+"))
      f->mode = VFS_WRITE | VFS_MUSTEXIST;
    if (strstr(mode, "w+"))
      f->mode = VFS_WRITE | VFS_CREATE | VFS_NEWCONTENT;
    if (strstr(mode, "a+"))
      f->mode = VFS_WRITE | VFS_CREATE | VFS_SEEKEOF;
  }
  if (f->mode == 0 || ((f->mode & VFS_READ) && (f->mode & VFS_WRITE))) {
    free(f);
    return NULL;
  }

#ifdef NDE_ALLOW_NONCACHED
  if (!f->cached) {
    f->rfile = fopen(fl, mode);
    if (f->rfile)
      f->filename = strdup(fl);
    else {
      free(f);
      return NULL;
    }
  return f;
  }
#endif

  if (f->mode & VFS_MUSTEXIST)
  {
#ifdef NDE_NOWIN32FILEIO
    if (access(fl, 0))
    {
      free(f);
      return NULL;
    }
#else
    HANDLE hFile = CreateFile(fl,0,0,0,OPEN_EXISTING,0,0);
    if (hFile == INVALID_HANDLE_VALUE)
    {
      free(f);
      return NULL;
    }
    CloseHandle(hFile);
#endif
  }

  if (!(f->mode & VFS_NEWCONTENT))
  {
#ifdef NDE_NOWIN32FILEIO
    FILE *pf;
    pf = fopen(fl, "rb");
    if (!pf) {
      f->data = (unsigned char *)calloc(VFILE_INC, 1);
      if (f->data == NULL) {
        free(f);
        return NULL;
      }
      f->filesize = 0;
      f->maxsize = VFILE_INC;
    } else {
      fseek(pf, 0, SEEK_END);
      f->filesize = ftell(pf);
      fseek(pf, 0, SEEK_SET);
      f->data = (unsigned char *)calloc(f->filesize, 1);
      if (f->data == NULL) {
        free(f);
        return NULL;
      }
      f->maxsize = f->filesize;
      fread(f->data, f->filesize, 1, pf);
      fclose(pf);
    }
#else
    int retry_cnt=0;
    HANDLE hFile;
    hFile=CreateFile(fl,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_FLAG_SEQUENTIAL_SCAN,0);
    if (hFile==INVALID_HANDLE_VALUE)
    {
      f->data = (unsigned char *)calloc(VFILE_INC, 1);
      if (f->data == NULL) {
        free(f);
        return NULL;
      }
      f->filesize = 0;
      f->maxsize = VFILE_INC;
    }
    else
    {
      f->filesize = (int)GetFileSize(hFile,NULL);
      f->data = (unsigned char *)calloc(f->filesize, 1);
      if (f->data == NULL) {
        CloseHandle(hFile);
        free(f);
        return NULL;
      }
      f->maxsize = f->filesize;
      DWORD r;
      if (!ReadFile(hFile,f->data,f->filesize,&r,NULL) || r != f->filesize)
      {
        CloseHandle(hFile);
        free(f->data);
        free(f);
        return NULL;
      }
      CloseHandle(hFile);
    }
#endif
  }

  if (f->mode & VFS_SEEKEOF)
    f->ptr = f->filesize;

  f->filename = strdup(fl);
  return f;
}

//----------------------------------------------------------------------------
void Vfclose(VFILE *f) {
  if (!f) return;

#ifdef NDE_ALLOW_NONCACHED
  if (!f->cached) {
    free(f->filename);
    if (f->rfile) fclose(f->rfile);
    free(f);
    return;
  }
#endif

  if (!(f->mode & VFS_WRITE)) {
    free(f->filename);
    free(f->data);
    free(f);
    return;
  }

  Vsync(f);

  free(f->filename);
  free(f->data);
  free(f);
}

//----------------------------------------------------------------------------
int Vfread(void *ptr, size_t size, size_t n, VFILE *f) {
  if (!ptr || !f) return 0;
#ifdef NDE_ALLOW_NONCACHED
  if (!f->cached) {
    return fread(ptr, size, n, f->rfile);
  }
#endif
  int s = size*n;
  if (!s) return 0;
  if (s + f->ptr > f->filesize) {
    //FUCKO: remove this
    if(! (f->ptr < f->filesize))
    {
      char buf[1024];
      wsprintf(buf,"NDE/VFS: VFS read at %d/%d (%d bytes) is bad\n",f->ptr,f->filesize,s);
      OutputDebugString(buf);
      extern HWND g_hwnd;
//      if (!f->flushtable) // this would be ideal, if we could figure out f->flushtable
  //      f->flushtable=MessageBox(g_hwnd,"DB read failed, DB may be corrupted.\r\n\r\n"
    //    "Hit Retry to continue, or Cancel to clear the DB and start over","Winamp Library Error",MB_RETRYCANCEL) == IDCANCEL; //fucko
      //MessageBox(g_hwnd,"DB read failed, DB may be corrupted. If this error persists, remove all files from the library.",
//        "Winamp Library Error",MB_OK);
      return 0;
    }
    s = f->filesize - f->ptr;
  }
  memcpy(ptr, f->data+f->ptr, s);
  f->ptr += s;
  return s/size;
}

//----------------------------------------------------------------------------
void Vfwrite(void *ptr, size_t size, size_t n, VFILE *f) {
  if (!ptr || !f) return;
#ifdef NDE_ALLOW_NONCACHED
  if (!f->cached) {
    fwrite(ptr, size, n, f->rfile);
    return;
  }
#endif
  f->dirty=1;
  int s = size*n;
  if (s + f->ptr > f->maxsize) {
    // grow f->data,f->maxsize to be (s + f->ptr + VFILE_INC-1)&~(VFILE_INC-1)
    // instead of calling Vgrow again which gets kinda slow
    int newsize=(s + f->ptr + VFILE_INC-1)&~(VFILE_INC-1);
    f->data=(unsigned char *)realloc(f->data,newsize);
    if (f->data == NULL) return;
    memset(f->data+f->maxsize,0,newsize-f->maxsize);
    f->maxsize=newsize;
  }
  memcpy(f->data + f->ptr, ptr, s);
  f->ptr += s;
  if (f->ptr > f->filesize)
    f->filesize = f->ptr;
}

//----------------------------------------------------------------------------
void Vgrow(VFILE *f) {
  if (!f) return;
#ifdef NDE_ALLOW_NONCACHED
  if (!f->cached) return;
#endif
  unsigned char *newBlock = (unsigned char *)calloc(f->maxsize + VFILE_INC, 1);
  if (f->data)
    memcpy(newBlock, f->data, f->maxsize);
  f->maxsize += VFILE_INC;
  if (f->data)
    free(f->data);
  f->data = newBlock;
}

//----------------------------------------------------------------------------
void Vfputs(char *str, VFILE *f) {
  if (!f) return;
#ifdef NDE_ALLOW_NONCACHED
  if (!f->cached) {
    fputs(str, f->rfile);
    return;
  }
#endif
  Vfwrite(str, strlen(str), 1, f);
}

//----------------------------------------------------------------------------
void Vfputc(char c, VFILE *f) {
  if (!f) return;
#ifdef NDE_ALLOW_NONCACHED
  if (!f->cached) {
    fputc(c, f->rfile);
    return;
  }
#endif
  Vfwrite(&c, 1, 1, f);
}

// not mac comliant
//----------------------------------------------------------------------------
char *Vfgets(char *dest, int n, VFILE *f) {
  if (!f) return NULL;
#ifdef NDE_ALLOW_NONCACHED
  if (!f->cached)
    return fgets(dest, n, f->rfile);
#endif

  unsigned char c=0;
  char *p;
  int l=0;

  p = dest;
  while (l < n && !Vfeof(f)) {
    c = f->data[f->ptr];
    f->ptr++;
    *p = c;
    p++;
    l++;
    if (c == '\n') {
      if (!Vfeof(f) && f->data[f->ptr] == '\r') {
        f->ptr++;
      }
    break;
    }
  }
  *p=0;
  return dest;
}

//----------------------------------------------------------------------------
char Vfgetc(VFILE *f) {
  if (!f) return EOF;
#ifdef NDE_ALLOW_NONCACHED
  if (!f->cached)
    return fgetc(f->rfile);
#endif
  if (!Vfeof(f))
    return f->data[f->ptr++];
  return EOF;
}

//----------------------------------------------------------------------------
unsigned long Vftell(VFILE *f) {
  if (!f) return (unsigned)-1;
#ifdef NDE_ALLOW_NONCACHED
  if (!f->cached)
    return ftell(f->rfile);
#endif
  return f->ptr;
}

//----------------------------------------------------------------------------
void Vfseek(VFILE *f, long i, int whence) {
  if (!f) return;
#ifdef NDE_ALLOW_NONCACHED
  if (!f->cached) {
    fseek(f->rfile, i, whence);
    return;
  }
#endif
  switch (whence) {
    case SEEK_SET:
      f->ptr = i;
      break;
    case SEEK_CUR:
      f->ptr += i;
      break;
    case SEEK_END:
      f->ptr = f->filesize+i;
      break;
  }
}

//----------------------------------------------------------------------------
int Vfeof(VFILE *f) {
  if (!f) return -1;
#ifdef NDE_ALLOW_NONCACHED
  if (!f->cached) return feof(f->rfile);
#endif
  return (f->ptr >= f->filesize);
}

//----------------------------------------------------------------------------
int Vsync(VFILE *f) {
  if (!f) return 0;
  if (!f->dirty) return 0;

  if (f->mode & VFS_WRITE)
  {
#ifdef NDE_ALLOW_NONCACHED
    if (!f->cached)
    {
      int p=ftell(f->rfile);
      fclose(f->rfile);
      f->rfile = fopen(f->filename, "r+b");
      if (!f->rfile)
        return 0;
      fseek(f->rfile, p, SEEK_SET);
      return 0;
    }
#endif

    char *newfn=(char*)malloc(strlen(f->filename)+32);
    char *oldfn=(char*)malloc(strlen(f->filename)+32);
    DWORD mypid=GetCurrentProcessId();

    wsprintf(newfn,"%s.n3w%08X",f->filename,mypid);
    wsprintf(oldfn,"%s.o1d%08X",f->filename,mypid);
#ifdef NDE_NOWIN32FILEIO
    unlink(newfn);
    unlink(oldfn);
#else
    DeleteFile(newfn);
    DeleteFile(oldfn);
#endif

tryagain:


#ifdef NDE_NOWIN32FILEIO
    FILE *pf = fopen(newfn, "wb");
    if (!pf || fwrite(f->data, f->filesize, 1, pf) != 1)
    {
      if (pf) fclose(pf);

      #ifdef WIN32
      extern HWND g_hwnd;
      int a=MessageBox(g_hwnd,"Error flushing DB to disk (is your drive full?).\r\n\r\n"
        "Hit Retry to try again (recommended), or Cancel to abort (NOT recommended, and may leave the DB in an unuseable state).\r\n",
        "Winamp Library Error",MB_RETRYCANCEL);
      if (a == IDRETRY) goto tryagain;
      #else
      printf("Error flushing DB to disk (is your drive full?)\n\nHit (R)etry to try again (recommended), or (C)ancel to abort (NOT recommended, and may leave the DB in an unuseable state)");
      fflush(stdout);
      char c;
      while (1) {
        scanf("%c", &c);
        clear_stdin();
        c = toupper(c);
        if (c == 'R') goto tryagain;
        else if (c == 'C') break;
      }
      #endif

      unlink(newfn);
      free(newfn);
      free(oldfn);
      return 1;
    }
    fclose(pf);
    rename(f->filename,oldfn); // save old file
#else
    HANDLE hFile = CreateFile(newfn,GENERIC_WRITE,0,0,CREATE_ALWAYS,0,NULL);
    int success=0;
    if (hFile != INVALID_HANDLE_VALUE)
    {
      DWORD o;
      if (WriteFile(hFile,f->data,f->filesize,&o,NULL) && o == f->filesize) success++;
      CloseHandle(hFile);
    }
    if (!success)
    {

      #ifdef WIN32
      extern HWND g_hwnd;
      int a=MessageBox(g_hwnd,"Error flushing DB to disk (is your drive full?).\r\n\r\n"
        "Hit Retry to try again (recommended), or Cancel to abort (NOT recommended, and may leave the DB in an unuseable state).\r\n",
        "Winamp Library Error",MB_RETRYCANCEL);
      if (a == IDRETRY) goto tryagain;

      #else
      printf("Error flushing DB to disk (is your drive full?)\n\nHit (R)etry to try again (recommended), or (C)ancel to abort (NOT recommended, and may leave the DB in an unuseable state)");
      fflush(stdout);
      char c;
      while (1) {
        scanf("%c", &c);
        clear_stdin();
        c = toupper(c);
        if (c == 'R') goto tryagain;
        else if (c == 'C') break;
      }
      #endif

      DeleteFile(newfn);
      free(newfn);
      free(oldfn);
      return 1;
    }

    MoveFile(f->filename,oldfn); // save old file
#endif


    int rv=0;

#ifdef NDE_NOWIN32FILEIO
  tryagain2:
    if (rename(newfn,f->filename))
    {
      #ifdef WIN32
      extern HWND g_hwnd;
      int a=MessageBox(g_hwnd,"Error updating DB file on disk. This should never really happen.\r\n\r\n"
        "Hit Retry to try again (recommended), or Cancel to abort (NOT recommended, and may leave the DB in an unuseable state).\r\n",
        "Winamp Library Error",MB_RETRYCANCEL);
      if (a == IDRETRY) goto tryagain2;

      #else
      printf("Error updating DB file on disk. This should never really happen\n\nHit (R)etry to try again (recommended), or (C)ancel to abort (NOT recommended, and may leave the DB in an unuseable state)");
      fflush(stdout);
      char c;
      while (1) {
        scanf("%c", &c);
        clear_stdin();
        c = toupper(c);
        if (c == 'R') goto tryagain2;
        else if (c == 'C') break;
      }
      #endif

      rename(oldfn,f->filename); // restore old file
      rv=1;
    }

    // clean up our temp files
    unlink(oldfn);
    unlink(newfn);

#else
  tryagain2:
    if (!MoveFile(newfn,f->filename))
    {
      #ifdef WIN32
      extern HWND g_hwnd;
      int a=MessageBox(g_hwnd,"Error updating DB file on disk. This should never really happen.\r\n\r\n"
        "Hit Retry to try again (recommended), or Cancel to abort (NOT recommended, and may leave the DB in an unuseable state).\r\n",
        "Winamp Library Error",MB_RETRYCANCEL);
      if (a == IDRETRY) goto tryagain2;

      #else
      printf("Error updating DB file on disk. This should never really happen\n\nHit (R)etry to try again (recommended), or (C)ancel to abort (NOT recommended, and may leave the DB in an unuseable state)");
      fflush(stdout);
      char c;
      while (1) {
        scanf("%c", &c);
        clear_stdin();
        c = toupper(c);
        if (c == 'R') goto tryagain2;
        else if (c == 'C') break;
      }
      #endif

      MoveFile(oldfn,f->filename); // restore old file
      rv=1;
    }

    // clean up our temp files
    DeleteFile(oldfn);
    DeleteFile(newfn);
#endif

    free(newfn);
    free(oldfn);
    return rv;
  }
  f->dirty=0;
  return 0;
}


