#ifndef __VFS_H
#define __VFS_H

#ifdef WIN32
#include <windows.h>
#endif

#ifdef NDE_NOWIN32FILEIO
#include <stdio.h>
#endif

#ifdef NDE_ALLOW_NONCACHED
  #ifndef NDE_NOWIN32FILEIO
    #error NDE_ALLOW_NONCACHED at least for now requires NDE_NOWIN32FILEIO
  #endif
#endif

#define VFILE_INC 65536

#define VFS_READ        1
#define VFS_WRITE        2
#define VFS_SEEKEOF      4
#define VFS_CREATE      8
#define VFS_NEWCONTENT 16
#define VFS_MUSTEXIST  32

#ifndef _SIZE_T
#define _SIZE_T
typedef unsigned size_t;
#endif

typedef struct {
    unsigned char *data;
    unsigned long ptr;
    unsigned long filesize;
    unsigned long maxsize;
    char *filename;
    char mode;
#ifdef NDE_ALLOW_NONCACHED
    BOOL cached;
#endif
    int dirty;
    int flushtable;

#ifdef NDE_ALLOW_NONCACHED
  #ifdef NDE_NOWIN32FILEIO
    FILE *rfile;
  #else
    HANDLE hfile;
  #endif
#endif

} VFILE;

#include "NDE.h"

#ifdef __cplusplus
extern "C" {
#endif

VFILE *Vfopen(char *fl, char *mode, BOOL Cached);
int Vfread(void *ptr, size_t size, size_t n, VFILE *buf);
void Vfseek(VFILE *fl, long i, int whence);
unsigned long Vftell(VFILE *fl);
void Vfclose(VFILE *fl);
void Vfwrite(void *ptr, size_t size, size_t n, VFILE *f);
void Vfputs(char *str, VFILE *fl);
void Vgrow(VFILE *f);
void Vfputc(char c, VFILE *fl);
int Vfeof(VFILE *fl);
char *Vfgets(char *dest, int n, VFILE *fl);
char Vfgetc(VFILE *fl);
int Vsync(VFILE *fl); // 1 on error updating

#ifdef __cplusplus
}
#endif

#endif

